@@ -45,7 +45,7 @@ PERFORMANCE OF THIS SOFTWARE.
  *
  */
 
-static char __version__[] = "1.2";
+static char __version__[] = "1.3";
 
 static char pymqe_doc[] = " \
 pymqe - A Python MQ Extension.  This presents a low-level Python \
@@ -64,7 +64,7 @@ elements of a tuple.\
 \
 The MQ verbs implemented here are:\
   MQCONN, MQCONNX, MQDISC, MQOPEN, MQCLOSE, MQPUT, MQPUT1, MQGET,\
-  MQCMIT, MQBACK, MQBEGIN, MQINQ, MQSET\
+  MQCMIT, MQBACK, MQBEGIN, MQINQ, MQSET, MQSUB, MQCRTMH, MQSETMP, MQINQMP\
 \
 The PCF MQAI call mqExecute is also implemented.\
 \
@@ -112,8 +112,6 @@ static PyObject *ErrorObj;
  * MQI Structure sizes for the current supported MQ version are
  * defined here for convenience. This allows older versions of pymqi
  * to work with newer versions of MQI.
- *
- * Current supported MQ Versions are 5.2. Known to work with 5.3
  */
 
 #define PYMQI_MQCD_SIZEOF MQCD_CURRENT_LENGTH
@@ -127,8 +125,21 @@ static PyObject *ErrorObj;
 #ifdef MQCMDL_LEVEL_700
 #define PYMQI_MQSD_SIZEOF sizeof(MQSD)
 #define PYMQI_MQSRO_SIZEOF sizeof(MQSRO)
+#define PYMQI_MQCMHO_SIZEOF sizeof(MQCMHO)
+#define PYMQI_MQSMPO_SIZEOF sizeof(MQSMPO)
+#define PYMQI_MQIMPO_SIZEOF sizeof(MQIMPO)
+#define PYMQI_MQPD_SIZEOF sizeof(MQPD)
 #endif
 
+/* Macro for cleaning up MQAI filters-related object.
+*/
+#define PYMQI_MQAI_FILTERS_CLEANUP \
+  Py_XDECREF(filter_selector); \
+  Py_XDECREF(filter_operator); \
+  Py_XDECREF(filter_value); \
+  Py_XDECREF(_pymqi_filter_type); \
+  Py_XDECREF(filter);
+
 
 /* ----------------------------------------------------- */
 
@@ -758,6 +769,166 @@ static PyObject * pymqe_MQSUB(PyObject *self, PyObject *args) {
 
 #ifdef PYMQI_FEATURE_MQAI
 
+/* Message properties and selectors - start */
+
+#ifdef MQCMDL_LEVEL_700
+
+static char pymqe_MQCRTMH__doc__[] =
+"MQCRTMH(conn_handle, cmho) \
+ \
+Calls the MQI's MQCRTMH function \
+";
+
+static PyObject* pymqe_MQCRTMH(PyObject *self, PyObject *args) {
+
+  long conn_handle;
+
+  char *cmho_buffer;
+  long cmho_buffer_length;
+
+  MQCMHO *cmho;
+  MQHMSG msg_handle = MQHM_UNUSABLE_HMSG;
+  MQLONG comp_code = MQCC_UNKNOWN, comp_reason = MQRC_NONE;
+
+  PyObject *rv;
+
+  if (!PyArg_ParseTuple(args, "ls#", &conn_handle, &cmho_buffer, &cmho_buffer_length)) {
+    return NULL;
+  }
+
+  if (checkArgSize(cmho_buffer_length, PYMQI_MQCMHO_SIZEOF, "MQCMHO")) {
+    return NULL;
+  }
+
+  cmho = (MQCMHO *)cmho_buffer;
+
+  Py_BEGIN_ALLOW_THREADS
+  MQCRTMH(conn_handle, cmho, &msg_handle, &comp_code, &comp_reason);
+  Py_END_ALLOW_THREADS
+
+  rv = Py_BuildValue("(lll)", (long)comp_code, (long)comp_reason, (long)msg_handle);
+
+  return rv;
+}
+
+static char pymqe_MQSETMP__doc__[] =
+"MQSETMP(conn_handle, msg_handle, smpo, name, pd, type) \
+ \
+Calls the MQI's MQSETMP function \
+";
+
+static PyObject* pymqe_MQSETMP(PyObject *self, PyObject *args) {
+
+  long conn_handle, msg_handle;
+
+  char *smpo_buffer;
+  long smpo_buffer_length;
+
+  char *pd_buffer;
+  long pd_buffer_length;
+
+  MQSMPO *smpo;
+  MQPD *pd;
+
+  MQCHARV name = {MQCHARV_DEFAULT};
+  char *property_name;
+  long property_name_length = 0;
+
+  long property_type;
+
+  char *property_value;
+  long value_length;
+
+  MQLONG comp_code = MQCC_UNKNOWN, comp_reason = MQRC_NONE;
+  PyObject *rv;
+
+  if (!PyArg_ParseTuple(args, "lls#s#s#lsl", &conn_handle, &msg_handle, &smpo_buffer,
+                        &smpo_buffer_length,
+                        &property_name, &property_name_length,
+                        &pd_buffer, &pd_buffer_length, &property_type,
+                        &property_value, &value_length)) {
+    return NULL;
+  }
+
+  if (checkArgSize(smpo_buffer_length, PYMQI_MQSMPO_SIZEOF, "MQSMPO")) {
+    return NULL;
+  }
+
+  if (checkArgSize(pd_buffer_length, PYMQI_MQPD_SIZEOF, "MQPD")) {
+    return NULL;
+  }
+
+  smpo = (MQSMPO *)smpo_buffer;
+  pd = (MQPD *)pd_buffer;
+
+  name.VSPtr = property_name;
+  name.VSLength = property_name_length;
+
+  Py_BEGIN_ALLOW_THREADS
+  MQSETMP(conn_handle, msg_handle, smpo, &name, pd, property_type, value_length,
+            (MQBYTE *)property_value, &comp_code, &comp_reason);
+  Py_END_ALLOW_THREADS
+
+  rv = Py_BuildValue("(ll)", (long)comp_code, (long)comp_reason);
+  return rv;
+
+}
+
+static char pymqe_MQINQMP__doc__[] =
+"MQINQMP(conn_handle, msg_handle, smpo, name, pd, type, max_value_length) \
+ \
+Calls the MQI's MQINQMP function \
+";
+
+static PyObject* pymqe_MQINQMP(PyObject *self, PyObject *args) {
+
+  long conn_handle, msg_handle;
+  long impo_options, pd_options;
+
+  MQCHARV name = {MQCHARV_DEFAULT};
+  char *property_name;
+  long property_name_length = 0;
+
+  MQLONG property_type;
+  MQLONG actual_value_length;
+  long max_value_length;
+
+  MQLONG comp_code = MQCC_UNKNOWN, comp_reason = MQRC_NONE;
+  PyObject *rv;
+  
+  MQBYTE *value;
+  MQIMPO impo = {MQIMPO_DEFAULT};
+  MQPD pd = {MQPD_DEFAULT};
+
+  if (!PyArg_ParseTuple(args, "llls#lll", &conn_handle, &msg_handle, &impo_options,
+                        &property_name, &property_name_length, &pd_options,
+                        &property_type, &max_value_length)) {
+    return NULL;
+  }
+
+  value = malloc(max_value_length * sizeof(MQBYTE));
+  
+  impo.Options = impo_options;
+  pd.Options = pd_options;
+
+  name.VSPtr = property_name;
+  name.VSLength = property_name_length;
+
+  MQINQMP(conn_handle, msg_handle, &impo, &name, &pd, &property_type, sizeof(value),
+    value, &actual_value_length, &comp_code, &comp_reason);
+
+  rv = Py_BuildValue("(lls)", (long)comp_code, (long)comp_reason, (char *)value);
+  free(value);
+  
+  return rv;
+
+}
+
+#endif /* MQCMDL_LEVEL_700 */
+
+/* Message properties and selectors - end */
+
+
 
 static void cleanupBags(MQHBAG adminBag, MQHBAG responseBag) {
   MQLONG compCode, compReason;
@@ -797,16 +968,31 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
     PyObject *returnValue;
     PyObject *resultDictValue;
     PyObject *newList;
-
+    
     long lQmgrHandle, lCmdCode;
     int isByteString = 0;
+    
+    /* Filters */
+    PyObject *filters = NULL;
+    PyObject *filter = NULL;
+    PyObject *filter_selector = NULL;
+    PyObject *filter_value = NULL;
+    PyObject *filter_operator = NULL;
+    PyObject *_pymqi_filter_type = NULL;
+    char *filter_type = NULL;
+    int filter_keys_size = 0;
 
-    if(!PyArg_ParseTuple(args, "ll|O", &lQmgrHandle, &lCmdCode, &argDict)) {
+    if(!PyArg_ParseTuple(args, "ll|OO", &lQmgrHandle, &lCmdCode, &argDict, &filters)) {
         return NULL;
     }
 
     if(argDict && !PyDict_Check(argDict)) {
-        PyErr_SetString(ErrorObj, "args is not a dictionary");
+        PyErr_SetString(ErrorObj, "'argDict' is not a dictionary");
+        return NULL;
+    }
+    
+    if(filters && !PyList_Check(filters)) {
+        PyErr_SetString(ErrorObj, "'filters' is not a list");
         return NULL;
     }
 
@@ -830,7 +1016,106 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
         if (compCode != MQCC_OK) {
             break;
         }
+        
+        /* 
+         * MQAI filters.
+         */
+        if(filters) {
+          
+          int i;
+          filter_keys_size = (int)PyList_Size(filters);
+          
+          for(i = 0; i < filter_keys_size; i++) {
+            filter = PyList_GetItem(filters, i); /* Borrowed ref */
+            if(NULL == filter) {
+              PyErr_Format(ErrorObj, "'filter' object is NULL.");
+              PYMQI_MQAI_FILTERS_CLEANUP
+              return NULL;
+            }
+            
+            filter_selector = PyObject_GetAttrString(filter, "selector"); /* Owned ref */
+            if(NULL == filter_selector) {
+              PyErr_Format(ErrorObj, "'filter_selector' object is NULL.");
+              PYMQI_MQAI_FILTERS_CLEANUP
+              return NULL;
+            }
+              
+            filter_operator = PyObject_GetAttrString(filter, "operator"); /* Owned ref */
+            if(NULL == filter_operator) {
+              PyErr_Format(ErrorObj, "'filter_operator' object is NULL.");
+              PYMQI_MQAI_FILTERS_CLEANUP
+              return NULL;
+            }
+              
+            filter_value = PyObject_GetAttrString(filter, "value"); /* Owned ref */
+            if(NULL == filter_value) {
+              PyErr_Format(ErrorObj, "'filter_value' object is NULL.");
+              PYMQI_MQAI_FILTERS_CLEANUP
+              return NULL;
+            }
+              
+            _pymqi_filter_type = PyObject_GetAttrString(filter, "_pymqi_filter_type"); /* Owned ref */
+            if(NULL == _pymqi_filter_type) {
+              PyErr_Format(ErrorObj, "'_pymqi_filter_type' object is NULL.");
+              PYMQI_MQAI_FILTERS_CLEANUP
+              return NULL;
+            }
+              
+            filter_type = PyString_AsString(_pymqi_filter_type); 
 
+            /* String filter */
+            if(0 == strcmp(filter_type, "string")) {
+
+              mqAddStringFilter(adminBag, 
+                                  (MQLONG)PyLong_AsLong(filter_selector), 
+                                  (MQLONG)PyObject_Length(filter_value),
+                                  (PMQCHAR)PyString_AsString(filter_value), 
+                                  (MQLONG)PyLong_AsLong(filter_operator),
+                                  &compCode, 
+                                  &compReason); 
+                                  
+              if(!compCode == MQCC_OK) {
+                PyErr_Format(ErrorObj, "Could not invoke 'mqAddStringFilter' compCode=[%d], " \
+                            "compReason=[%d], filter_selector=[%d], filter_value=[%s], " \
+                            "filter_operator=[%d]", (int)compCode, (int)compReason, (int)PyLong_AsLong(filter_selector), 
+                                  PyString_AsString(filter_value), 
+                                  (int)PyLong_AsLong(filter_operator));
+                                  
+                PYMQI_MQAI_FILTERS_CLEANUP
+                return NULL;
+              }
+                                  
+              
+            }
+            
+            /* Integer filter */
+            else if(0 == strcmp(filter_type, "integer")) {
+              
+              mqAddIntegerFilter(adminBag, (MQLONG)PyLong_AsLong(filter_selector), 
+                                  (MQLONG)PyLong_AsLong(filter_value), 
+                                  (MQLONG)PyLong_AsLong(filter_operator), &compCode, &compReason); 
+                                  
+              if(!compCode == MQCC_OK) {
+                PyErr_Format(ErrorObj, "Could not invoke 'mqAddIntegerFilter' compCode=[%d], " \
+                            "compReason=[%d], filter_selector=[%d], filter_value=[%d], " \
+                            "filter_operator=[%d]", (int)compCode, (int)compReason, (int)PyLong_AsLong(filter_selector), 
+                                  (int)PyLong_AsLong(filter_value), 
+                                  (int)PyLong_AsLong(filter_operator));
+                                  
+                PYMQI_MQAI_FILTERS_CLEANUP
+                return NULL;
+              }
+            }
+            else {
+              PyErr_Format(ErrorObj, "Unrecognized filter type [%s].", filter_type);
+              PYMQI_MQAI_FILTERS_CLEANUP
+              return NULL;
+            }
+          }
+          /* All's good, let's clean up after ourselves. */
+          PYMQI_MQAI_FILTERS_CLEANUP
+        }
+        
         /*
          * For each arg key/value pair, create the appopriate type and add
          * it to the bag.
@@ -852,7 +1137,7 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
                   PyObject *value = PyDict_GetItem(argDict, key);  /* Borrowed ref */
 
                   /*
-                   * The key ought to be a int or long. Blow up if it isn't
+                   * The key ought to be an int or a long. Blow up if it isn't
                    */
                   if(!PyLong_Check(key) && !PyInt_Check(key)) {
                       PyObject *keyStr = PyObject_Str(key);  /* Owned ref */
@@ -888,7 +1173,7 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
                         strByteArg = (MQBYTE *)PyString_AsString(byteStringValue);
 
 #ifdef MQCMDL_LEVEL_700
-                        mqAddByteString(adminBag, paramType, PyObject_Length(value), strByteArg, &compCode, &compReason);
+                        mqAddByteString(adminBag, paramType, (MQLONG)PyObject_Length(value), strByteArg, &compCode, &compReason);
 #endif /* MQCMDL_LEVEL_700 */
 
                         Py_XDECREF(byteStringValue);
@@ -1216,6 +1501,9 @@ static struct PyMethodDef pymqe_methods[] = {
 #endif
 #ifdef MQCMDL_LEVEL_700
   {"MQSUB", (PyCFunction)pymqe_MQSUB, METH_VARARGS, pymqe_MQSUB__doc__},
+  {"MQCRTMH", (PyCFunction)pymqe_MQCRTMH, METH_VARARGS, pymqe_MQCRTMH__doc__},
+  {"MQSETMP", (PyCFunction)pymqe_MQSETMP, METH_VARARGS, pymqe_MQSETMP__doc__},
+  {"MQINQMP", (PyCFunction)pymqe_MQINQMP, METH_VARARGS, pymqe_MQINQMP__doc__},
 #endif
   {NULL, (PyCFunction)NULL, 0, NULL}        /* sentinel */
 };
@@ -1269,6 +1557,18 @@ __declspec(dllexport)
 #endif
 #ifdef MQCMDL_LEVEL_700
       PyList_Append(versions, PyString_FromString("7.0"));
+#endif
+#ifdef MQCMDL_LEVEL_710
+      PyList_Append(versions, PyString_FromString("7.1"));
+#endif
+#ifdef MQCMDL_LEVEL_750
+      PyList_Append(versions, PyString_FromString("7.5"));
+#endif
+#ifdef MQCMDL_LEVEL_710
+      PyList_Append(versions, PyString_FromString("7.1"));
+#endif
+#ifdef MQCMDL_LEVEL_750
+      PyList_Append(versions, PyString_FromString("7.5"));
 #endif
       PyDict_SetItemString(d,"__mqlevels__", PyList_AsTuple(versions));
       Py_XDECREF(versions);
