@@ -25,16 +25,21 @@ The classes are:
 
     * MQOpts - Base class for defining MQI parameter structures.
     * CD - MQI MQCD structure class
+    * CMHO - MQI MQCMHO structure class
     * MD - MQI MQMD structure class
     * GMO - MQI MQGMO structure class.
+    * IMPO - MQI MQIMPO structure class
     * OD - MQI MQOD structure class.
+    * PD - MQI MQPD structure class.
     * PMO - MQI MQPMO structure class.
     * RFH2 - MQI MQRFH2 structure class.
     * SCO - MQI MQSCO structure class
+    * SMPO - MQI MQSMPO structure class
     * SRO - MQI MQSRO structure class
     * SD - MQI MQSD structure class
     * TM - MQI MQTM structure class
     * TMC2- MQI MQTMC2 structure class
+    * Filter/StringFilter/IntegerFilter - PCF/MQAI filters
     * QueueManager - Queue Manager operations
     * Queue - Queue operations
     * Topic - Publish/subscribe topic operations
@@ -103,7 +108,7 @@ except Exception:
 # PyMQI
 import pymqe, CMQC, CMQCFC, CMQXC
 
-__version__ = "1.2"
+__version__ = "1.3"
 __mqlevels__ = pymqe.__mqlevels__
 __mqbuild__ = pymqe.__mqbuild__
 
@@ -796,7 +801,11 @@ class cd(MQOpts):
     # until runtime so set it once here when first importing pymqi
     # (originally written by Brent S. Elmer, Ph.D. (mailto:webe3vt@aim.com)).
 
-    if '7.0' in pymqe.__mqlevels__:
+    if '7.5' in pymqe.__mqlevels__:
+        _mqcd_version = CMQXC.MQCD_VERSION_10
+        _mqcd_current_length = CMQXC.MQCD_LENGTH_10
+
+    elif '7.0' in pymqe.__mqlevels__:
         _mqcd_version = CMQXC.MQCD_VERSION_9
         _mqcd_current_length = CMQXC.MQCD_LENGTH_9
 
@@ -911,12 +920,18 @@ class cd(MQOpts):
                      ['MaxInstancesPerClient', 999999999, MQLONG_TYPE],
                      ['ClientChannelWeight', 0, MQLONG_TYPE],
                      ['ConnectionAffinity', 1, MQLONG_TYPE]]  # 1 = MQCAFTY_PREFERRED
+            
+        if '7.1' in pymqe.__mqlevels__:
+            opts += [['BatchDataLimit', 5000, MQLONG_TYPE],
+                     ['UseDLQ', 2, MQLONG_TYPE],
+                     ['DefReconnect', 0, MQLONG_TYPE]]
 
         # In theory, the pad should've been placed right before the 'MsgExitPtr'
         # attribute, however setting it there makes no effect and that's why
         # it's being set here, as a last element in the list.
-        if MQLONG_TYPE == 'i':
-            opts += [['pad','', '4s']]
+        if '7.1' not in pymqe.__mqlevels__:
+            if MQLONG_TYPE == 'i':
+                opts += [['pad','', '4s']]
 
         apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
@@ -1027,7 +1042,76 @@ class SRO(MQOpts):
 
         apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
+class CMHO(MQOpts):
+    """CMHO(**kw)
 
+    Construct an MQCMHO Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments
+    'kw'."""
+    def __init__(self, **kw):
+        opts = [['StrucId', CMQC.MQCMHO_STRUC_ID, '4s'],
+                ['Version', CMQC.MQCMHO_VERSION_1, MQLONG_TYPE],
+                ['Options', CMQC.MQCMHO_DEFAULT_VALIDATION, MQLONG_TYPE]]
+
+        apply(MQOpts.__init__, (self, tuple(opts)), kw)
+
+class PD(MQOpts):
+    """CMHO(**kw)
+
+    Construct an MQPD Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments
+    'kw'."""
+    def __init__(self, **kw):
+        opts = [['StrucId', CMQC.MQPD_STRUC_ID, '4s'],
+                ['Version', CMQC.MQPD_VERSION_1, MQLONG_TYPE],
+                ['Options', CMQC.MQPD_NONE, MQLONG_TYPE],
+                ['Support', CMQC.MQPD_SUPPORT_OPTIONAL, MQLONG_TYPE],
+                ['Context', CMQC.MQPD_NO_CONTEXT, MQLONG_TYPE],
+                ['CopyOptions', CMQC.MQCOPY_DEFAULT, MQLONG_TYPE]]
+
+        apply(MQOpts.__init__, (self, tuple(opts)), kw)
+
+class SMPO(MQOpts):
+    """SMPO(**kw)
+
+    Construct an MQSMPO Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments
+    'kw'."""
+    def __init__(self, **kw):
+        opts = [['StrucId', CMQC.MQSMPO_STRUC_ID, '4s'],
+                ['Version', CMQC.MQSMPO_VERSION_1, MQLONG_TYPE],
+                ['Options', CMQC.MQSMPO_SET_FIRST, MQLONG_TYPE],
+                ['ValueEncoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
+                ['ValueCCSID', CMQC.MQCCSI_APPL, MQLONG_TYPE]]
+
+        apply(MQOpts.__init__, (self, tuple(opts)), kw)
+
+class IMPO(MQOpts):
+    """IMPO(**kw)
+
+    Construct an MQIMPO Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments
+    'kw'."""
+    def __init__(self, **kw):
+        opts = [['StrucId', CMQC.MQIMPO_STRUC_ID, '4s'],
+                ['Version', CMQC.MQIMPO_VERSION_1, MQLONG_TYPE],
+                ['Options', CMQC.MQIMPO_INQ_FIRST, MQLONG_TYPE],
+                ['RequestedEncoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
+                ['RequestedCCSID', CMQC.MQCCSI_APPL, MQLONG_TYPE],
+                ['ReturnedEncoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
+                ['ReturnedCCSID', 0L, MQLONG_TYPE],
+                ['Reserved1', 0L, MQLONG_TYPE],
+
+                # ReturnedName
+                ['ReturnedNameVSPtr', 0, 'P'],
+                ['ReturnedNameVSOffset', 0L, MQLONG_TYPE],
+                ['ReturnedNameVSBufSize', 0L, MQLONG_TYPE],
+                ['ReturnedNameVSLength', 0L, MQLONG_TYPE],
+                ['ReturnedNameVSCCSID', 0L, MQLONG_TYPE],
+
+                ['TypeString', '', '8s']]
+
+        apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
 #
 # A utility to convert a MQ constant to its string mnemonic by groping
@@ -1172,6 +1256,7 @@ class QueueManager:
 
         Connect immediately to the Queue Manager 'name'."""
 
+        print(33333333, name)
         rv = pymqe.MQCONN(name)
         if rv[1]:
             raise MQMIError(rv[1], rv[2])
@@ -1711,7 +1796,7 @@ class Queue:
 
         return self.__qHandle
 
-#Publish Subscribe support - Hannes Wagener 201
+#Publish Subscribe support - Hannes Wagener 2011
 class Topic:
     """Topic(queue_manager, topic_name, topic_string, topic_desc, open_opts)
 
@@ -2048,7 +2133,183 @@ class Subscription:
                 self.close()
         except:
             pass
+            
+class MessageHandle(object):
+    """ A higher-level wrapper around the MQI's native MQCMHO structure.
+    """
 
+    # When accessing message properties, this will be the maximum number
+    # of characters a value will be able to hold. If it's not enough
+    # an exception will be raised and its 'actual_value_length' will be
+    # filled in with the information of how many characters there are actually
+    # so that an application may re-issue the call.
+    default_value_length = 64
+
+    class _Properties(object):
+        """ Encapsulates access to message properties.
+        """
+        def __init__(self, conn_handle, msg_handle):
+            self.conn_handle = conn_handle
+            self.msg_handle = msg_handle
+
+        def __getitem__(self, name):
+            """ Allows for a dict-like access to properties,
+            handle.properties[name]
+            """
+            value = self.get(name)
+            if not value:
+                raise KeyError('No such property [%s]' % name)
+
+            return value
+
+        def __setitem__(self, name, value):
+            """ Implements 'handle.properties[name] = value'.
+            """
+            return self.set(name, value)
+
+        def get(self, name, default=None, max_value_length=None,
+                impo_options=CMQC.MQIMPO_INQ_FIRST, pd=CMQC.MQPD_NONE,
+                property_type=CMQC.MQTYPE_AS_SET):
+            """ Returns the value of message property 'name'. 'default' is the
+            value to return if the property is missing. 'max_value_length'
+            is the maximum number of characters the underlying pymqe function
+            is allowed to allocate for fetching the value
+            (defaults to MessageHandle.default_value_length). 'impo_options'
+            and 'pd_options' describe options of MQPD and MQIMPO structures
+            to be used and 'property_type' points to the expected data type
+            of the property.
+            """
+            if not max_value_length:
+                max_value_length = MessageHandle.default_value_length
+
+            comp_code, comp_reason, value = pymqe.MQINQMP(self.conn_handle,
+                    self.msg_handle, impo_options, name, impo_options,
+                    property_type, max_value_length)
+
+            if comp_code != CMQC.MQCC_OK:
+                raise MQMIError(comp_code, comp_reason)
+
+            return value
+
+
+        def set(self, name, value, property_type=CMQC.MQTYPE_STRING,
+                value_length=CMQC.MQVL_NULL_TERMINATED, pd=None, smpo=None):
+            """ Allows for setting arbitrary properties of a message. 'name'
+            and 'value' are mandatory. All other parameters are OK to use as-is
+            if 'value' is a string. If it isn't a string, the 'property_type'
+            and 'value_length' should be set accordingly. For further
+            customization, you can also use 'pd' and 'smpo' parameters for
+            passing in MQPD and MQSMPO structures.
+            """
+            pd = pd if pd else PD()
+            smpo = smpo if smpo else SMPO()
+
+            comp_code, comp_reason = pymqe.MQSETMP(self.conn_handle,
+                    self.msg_handle, smpo.pack(), name, pd.pack(),
+                    property_type, value, value_length)
+
+            if comp_code != CMQC.MQCC_OK:
+                raise MQMIError(comp_code, comp_reason)
+
+    def __init__(self, qmgr=None, cmho=None):
+        self.conn_handle = qmgr.get_handle() if qmgr else CMQC.MQHO_NONE
+        cmho = cmho if cmho else CMHO()
+
+        comp_code, comp_reason, self.msg_handle = pymqe.MQCRTMH(self.conn_handle,
+                                                            cmho.pack())
+
+        if comp_code != CMQC.MQCC_OK:
+            raise MQMIError(comp_code, comp_reason)
+
+        self.properties = self._Properties(self.conn_handle, self.msg_handle)
+        
+class _Filter(object):
+    """ The base class for MQAI filters. The initializer expectes user to provide
+    the selector, value and the operator to use. For instance, the can be respectively
+    MQCA_Q_DESC, 'MY.QUEUE.*', MQCFOP_LIKE. Compare with the pymqi.Filter class.
+    """
+    _pymqi_filter_type = None
+    
+    def __init__(self, selector, value, operator):
+        self.selector = selector
+        self.value = value
+        self.operator = operator
+        
+    def __repr__(self):
+        msg = '<%s at %s %s:%s:%s>'
+        return msg % (self.__class__.__name__, hex(id(self)), self.selector, 
+                      self.value, self.operator)
+
+class StringFilter(_Filter):
+    """ A subclass of pymqi._Filter suitable for passing MQAI string filters around.
+    """
+    _pymqi_filter_type = 'string'
+    
+class IntegerFilter(_Filter):
+    """ A subclass of pymqi._Filter suitable for passing MQAI integer filters around.
+    """
+    _pymqi_filter_type = 'integer'
+    
+class FilterOperator(object):
+    """ Creates low-level filters basing on what's been provided in the high-level
+    pymqi.Filter object.
+    """
+    operator_mapping = {
+        'less': CMQCFC.MQCFOP_LESS,
+        'equal': CMQCFC.MQCFOP_EQUAL,
+        'greater': CMQCFC.MQCFOP_GREATER,
+        'not_less': CMQCFC.MQCFOP_NOT_LESS,
+        'not_equal': CMQCFC.MQCFOP_NOT_EQUAL,
+        'not_greater': CMQCFC.MQCFOP_NOT_GREATER,
+        'like': CMQCFC.MQCFOP_LIKE,
+        'not_like': CMQCFC.MQCFOP_NOT_LIKE,
+        'contains': CMQCFC.MQCFOP_CONTAINS,
+        'excludes': CMQCFC.MQCFOP_EXCLUDES,
+        'contains_gen': CMQCFC.MQCFOP_CONTAINS_GEN,
+        'excludes_gen': CMQCFC.MQCFOP_EXCLUDES_GEN,
+        }
+    
+    def __init__(self, pub_filter):
+        self.pub_filter = pub_filter
+        
+    def __call__(self, value):
+        
+        # Do we support the given attribute filter?
+        if self.pub_filter.selector >= CMQC.MQIA_FIRST and self.pub_filter.selector <= CMQC.MQIA_LAST:
+            priv_filter_class = IntegerFilter
+        elif self.pub_filter.selector >= CMQC.MQCA_FIRST and self.pub_filter.selector <= CMQC.MQCA_LAST:
+            priv_filter_class = StringFilter
+        else:
+            msg = 'selector [%s] is of an unsupported type (neither integer ' \
+                'nor a string attribute). Send details to http://packages.' \
+                'python.org/pymqi/support-consulting-contact.html'
+            raise Error(msg % self.pub_filter.selector)
+
+        # Do we support the operator?
+        operator = self.operator_mapping.get(self.pub_filter.operator)
+        if not operator:
+            msg = 'Operator [%s] is not supported.'
+            raise Error(msg % self.pub_filter.operator)
+
+        return priv_filter_class(self.pub_filter.selector, value, operator)
+    
+class Filter(object):
+    """ The user-facing MQAI filtering class which provides syntactic sugar
+    on top of pymqi._Filter and its base classes.
+    """
+    def __init__(self, selector):
+        self.selector = selector
+        self.operator = None
+        
+    def __getattribute__(self, name):
+        """ A generic method for either fetching the pymqi.Filter object's
+        attributes or calling magic methods like 'like', 'contains' etc.
+        """
+        if name in('selector', 'operator'):
+            return object.__getattribute__(self, name)
+        self.operator = name
+                
+        return FilterOperator(self)
 #
 # This piece of magic shamelessly plagiarised from xmlrpclib.py. It
 # works a bit like a C++ STL functor.
@@ -2069,7 +2330,7 @@ class _Method:
         else:
             qmHandle = self.__pcf.getHandle()
         if len(args):
-            rv = pymqe.mqaiExecute(qmHandle, CMQCFC.__dict__[self.__name], args[0])
+            rv = pymqe.mqaiExecute(qmHandle, CMQCFC.__dict__[self.__name], *args)
         else:
             rv = pymqe.mqaiExecute(qmHandle, CMQCFC.__dict__[self.__name])
         if rv[1]:
@@ -2171,107 +2432,5 @@ def connect(queue_manager, channel=None, conn_info=None):
         return qmgr
     
     elif queue_manager:
-        qmgr = pymqi.QueueManager(queue_manager)
+        qmgr = QueueManager(queue_manager)
         return qmgr
-
-#######################################################################
-#
-# TESTS
-# This test code puts & gets a message on a queue. The queue manager
-# is the default (''). With no arguments, the queue 'QTEST1' is
-# used. This example is elaborated to show the pymqi features. It also
-# messes about with inquire & set calls.
-#######################################################################
-
-
-if __name__ == '__main__':
-    print 'pymqi version:', __version__, 'pymqe version:', pymqe.__version__, pymqe.__mqbuild__
-    print 'MQI Version:', max(map(float, __mqlevels__))
-    import sys, string
-    if len(sys.argv) > 1:
-        qName = sys.argv[1]
-    else:
-        qName = 'QTEST1'
-    print 'Using qname', qName
-    if pymqe.__mqbuild__ == 'server':
-        # Defaults to queue manager named ''
-        qMgr = QueueManager()
-    else:
-        qMgr = QueueManager(None)
-        qMgr.connectTCPClient('les.queue.manager', cd(), "SYSTEM.DEF.SVRCONN", 'localhost(1414)')
-
-
-    # Queue ctor can take a string as a queue name. Open is deferred
-    # until the put() call.
-    putQ = Queue(qMgr, qName)
-
-    # Queue ctor can also take a fully formed object descriptor. This
-    # queue is opened immediately.
-    getQDesc = od(ObjectName=qName)
-    getQ = Queue(qMgr, getQDesc)
-
-    # Put a message. Because of the deferred open, the put queue is
-    # opened now. Default values for message descriptor & put message
-    # options are used.
-    putQ.put('Hello pymqi!')
-
-    # Now query the queue & the queue manager
-    print 'Queue Manager name:', string.strip(qMgr.inquire(CMQC.MQCA_Q_MGR_NAME)),
-    print 'last altered:', qMgr.inquire(CMQC.MQCA_ALTERATION_DATE),
-    print 'on:', qMgr.inquire(CMQC.MQCA_ALTERATION_TIME)
-    qmgrPlatform = qMgr.inquire(CMQC.MQIA_PLATFORM)
-    platformDict = {3 : 'Unix', 5 : 'Windows', 11 : 'Windows/NT'}
-    try:
-        print 'Queue Manager Platform:', platformDict[qmgrPlatform]
-    except:
-        print 'Queue Manager Platform type:', qmgrPlatform
-
-    inqQ = Queue(qMgr, qName)
-    print 'Queue Depth:', inqQ.inquire(CMQC.MQIA_CURRENT_Q_DEPTH)
-    print 'last altered:', inqQ.inquire(CMQC.MQCA_ALTERATION_DATE),
-    print 'on:', inqQ.inquire(CMQC.MQCA_ALTERATION_TIME)
-
-    # Now do a get
-    try:
-        # Default message descriptor
-        msgDesc = md()
-
-        # MQI structure members can be set using keywords...
-        getOpts = gmo(Options = CMQC.MQGMO_WAIT)
-        # ... or attributes
-        getOpts.WaitInterval = CMQC.MQWI_UNLIMITED
-        # Get message from queue.
-        msg = getQ.get(None, msgDesc, getOpts)
-
-        # Some structure parameters are updated by the MQI call.
-        # Structure members can also be set/get as dictionary items.
-        print "Read message '%s' from resolved queue %s" % (msg, getOpts['ResolvedQName'])
-        print 'Queue depth now:', inqQ.inquire(CMQC.MQIA_CURRENT_Q_DEPTH)
-
-        # Test set
-        setQ = Queue(qMgr, qName)
-        setQ.set(CMQC.MQIA_INHIBIT_GET, CMQC.MQQA_GET_INHIBITED)
-        print 'INHIBIT_GET Value:', inqQ.inquire(CMQC.MQIA_INHIBIT_GET)
-        setQ.set(CMQC.MQIA_INHIBIT_GET, CMQC.MQQA_GET_ALLOWED)
-        setQ.set(CMQC.MQCA_TRIGGER_DATA, '%-64s' % 'Hello from Python!')
-    # Catch & print any pymqi exception
-    except Error, e:
-        print 'Got error', e
-
-    # Test PCF, if its available
-    if "5.3" in pymqe.__mqlevels__:
-        print 'PCF tests'
-        # FIXME: Allow connection to be passed in
-        pcf = PCFExecute(qMgr)
-        print 'Pinged queue manager'
-        pcf.MQCMD_PING_Q_MGR()
-        print 'Querying Queue'
-        results = pcf.stringifyKeys(pcf.CMQCFC.MQCMD_INQUIRE_Q({CMQC.MQCA_Q_NAME : qName})[0])
-        print results
-        for x in results:
-            if type(x) is not types.StringType:
-                print 'PCF key %d (value %s) is not known' % (x, results[x])
-    else:
-        print 'PCF/MQAI not available'
-
-    print 'Test completed'
